from oba.oba import NotFound, Obj, raw, iterable, is_dict, is_list, is_tuple, BaseObj
from oba.path import Path
from oba.soba import Soba


def obj(object_):
    return Obj(object_)


def soba(object_, separator='.'):
    return Soba(object_, separator=separator)


__all__ = [BaseObj, NotFound, Obj, Path, Soba, obj, soba, raw, iterable, is_dict, is_list, is_tuple]
__version__ = '0.3.1'



class Path:
    def __init__(self, path=None, vi=None):
        self._path = path or []
        self._valid_index = vi

    def mark(self):
        if self._valid_index is None:
            self._valid_index = len(self._path) - 1

    def __truediv__(self, other: str):
        if not isinstance(other, str):
            raise ValueError("Path components must be strings")
        return Path(self._path.copy() + [other], vi=self._valid_index)

    def __call__(self):
        return '→'.join(self._path)

    def __len__(self):
        return len(self._path)

    def __str__(self):
        valid_path = self._path[:self._valid_index]
        invalid_path = self._path[self._valid_index:]
        valid_path = list(map(lambda x: f'✓{x}', valid_path))
        invalid_path = list(map(lambda x: f'✗{x}', invalid_path))
        return ' → '.join(valid_path + invalid_path)



import abc
import warnings

from oba.path import Path


def raw(object_: 'BaseObj'):
    if isinstance(object_, NotFound):
        return object.__getattribute__(object_, '__path__')()
    if isinstance(object_, Obj):
        return object.__getattribute__(object_, '__obj__')
    return object_


def iterable(object_) -> bool:
    return isinstance(object_, dict) or isinstance(object_, list) or isinstance(object_, tuple)


def is_dict(object_: 'BaseObj') -> bool:
    return isinstance(object_, Obj) and isinstance(raw(object_), dict)


def is_list(object_: 'BaseObj') -> bool:
    return isinstance(object_, Obj) and isinstance(raw(object_), list)


def is_tuple(object_: 'BaseObj') -> bool:
    return isinstance(object_, Obj) and isinstance(raw(object_), tuple)


class BaseObj(abc.ABC):
    def __init__(self, object_=None, path=None):
        object.__setattr__(self, '__path__', path or Path())
        object.__setattr__(self, '__obj__', object_ if object_ is not None else {})

    def __iter__(self):
        raise NotImplementedError

    @classmethod
    def raw(cls, object_: 'BaseObj'):
        warnings.warn(
            'Obj.raw is deprecated, use oba.raw instead', DeprecationWarning, stacklevel=2)
        return raw(object_)

    @classmethod
    def iterable(cls, object_):
        warnings.warn(
            'Obj.iterable is deprecated, use oba.iterable instead', DeprecationWarning, stacklevel=2)
        return iterable(object_)

    def __contains__(self, item):
        raise NotImplementedError

    def __getitem__(self, key):
        raise NotImplementedError

    def __setitem__(self, key, value):
        raise NotImplementedError

    def __delitem__(self, key):
        raise NotImplementedError

    def __getattr__(self, key):
        raise NotImplementedError

    def __setattr__(self, key, value):
        raise NotImplementedError

    def __delattr__(self, key):
        raise NotImplementedError

    def __bool__(self):
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError

    def __str__(self):
        raise NotImplementedError


class NotFound(BaseObj):
    def __init__(self, object_=None, path=None):
        super().__init__(object_, path)

        if not isinstance(path, Path):
            raise ValueError('path for NotFound class should be a Path object')

        if not len(path):
            raise ValueError('path should have at least one component')

        path.mark()

    def __iter__(self):
        return iter({})

    def __contains__(self, item):
        return False

    def __getitem__(self, key):
        return NotFound(path=self.__path__ / key)

    def __setitem__(self, key, value):
        pass

    def __delitem__(self, key):
        pass

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        pass

    def __delattr__(self, key):
        pass

    def __bool__(self):
        return False

    def __str__(self):
        raise ValueError(f'Path not exists: {self.__path__}')

    def __len__(self):
        return 0


NoneObj = NotFound


class Obj(BaseObj):
    def __init__(self, object_=None, path=None):
        super().__init__(object_, path)

        if not iterable(object_):
            raise TypeError('obj should be iterable (dict, list, or tuple)')

    def __getitem__(self, key):
        # noinspection PyBroadException
        try:
            value = self.__obj__.__getitem__(key)
        except Exception:
            return NotFound(path=self.__path__ / key)
        if iterable(value):
            value = Obj(value, path=self.__path__ / str(key))
        return value

    def __getattr__(self, key: str):
        return self[key]

    def __setitem__(self, key: str, value):
        if not is_dict(self):
            raise TypeError('__setitem__ and __setattr__ can only be used on a dict-like object')
        self.__obj__[key] = value

    def __setattr__(self, key, value):
        self[key] = value

    def __contains__(self, item):
        return item in self.__obj__

    def __iter__(self):
        if isinstance(self.__obj__, dict):
            for key in self.__obj__:
                yield key
        else:
            for i, item in enumerate(self.__obj__):
                if iterable(item):
                    yield Obj(item, path=self.__path__ / str(i))
                else:
                    yield item

    def __len__(self):
        return len(self.__obj__)

    def __call__(self):
        return self.__obj__

    def __bool__(self):
        return bool(self.__obj__)

    def __str__(self):
        return 'Obj()'



from nestify import nestify
from oba.oba import raw, NotFound, iterable

from oba import Obj


class Soba(Obj):
    def __init__(self, object_=None, path=None, separator='.'):
        object_ = nestify(object_, separator=separator) if object_ is not None else object_
        super().__init__(object_, path)
        object.__setattr__(self, '__separator__', separator)

    def __getitem__(self, key):
        if isinstance(key, str):
            key = key.split(self.__separator__, maxsplit=1)

        index = key[0] if isinstance(key, list) else key

        # noinspection PyBroadException
        try:
            value = self.__obj__.__getitem__(index)
        except Exception:
            return NotFound(path=self.__path__ / index)
        if iterable(value):
            value = Obj(value, path=self.__path__ / str(key))

        if isinstance(key, list) and len(key) > 1:
            value = value[key[1]]
        return value

    def __setitem__(self, key: str, value):
        key = key.split(self.__separator__, maxsplit=1)

        if len(key) == 1:
            obj = raw(self)
            obj[key[0]] = value
        else:
            self[key[0]][key[1]] = value

    def __str__(self):
        return 'Soba()'



from setuptools import setup, find_packages

from pathlib import Path
from oba import __version__ as version

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='oba',
    version=version,
    keywords=['dict', 'object'],
    description='make iter object easy to access (item to attr)',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT Licence',
    url='https://github.com/Jyonn/Oba',
    author='Jyonn Liu',
    author_email='i@6-79.cn',
    platforms='any',
    packages=find_packages(),
    install_requires=[
        'nestify'
    ],
)
